/**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001, Salvatore Isaja                                    *
 *                                                                        *
 * This is "dir.c" - Remove and create directories, change current        *
 *                   directory                                            *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/

/* TO DO: fat_chdir and fat_cwd */

#include "fat1.h"


/* Removes the specified directory, checking if it's empty and */
/* calling fat_unlink (unlink.c)                               */
/* Returns 0 on success, or a negative error code on failure.  */
/* This is a public driver function.                           */
int fat_rmdir(void *P, char *DirName, int UseLfn)
{
  tFatFile     *F;
  tFatVolume   *V = (tFatVolume *) P;
  tFatDirEntry  D;
  int           Res;
  int           Fd;

  /* First we open the directory to see if removing it is a valid operation */
  Fd = fat_open(V, DirName, FD32_OPEN_ACCESS_READ | FAT_OPEN_DIRECTORY |
                FD32_OPEN_EXIST_OPEN, FD32_ATTR_NONE, 0, NULL, UseLfn);
  if (Fd < 0) return Fd;
  F = &V->Files[Fd];
  
  /* Check if trying to remove the current directory */
  if (F->FirstCluster == V->CurDirCluster) return FD32_ERROR_RMDIR_CURRENT;

  /* Check if trying to remove the root directory */
  if (((V->FatType == FATTYPE_FAT32)
    && (F->FirstCluster == V->Bpb.BPB_RootClus))
    || (F->Mode & FAT_FMODE_ROOT)) return FD32_ERROR_ACCESS_DENIED;
    
  /* Check if the directory contains dot and dotdot as first two entries */
  Res = fat_read(V, Fd, (BYTE *) &D, 32, NULL);
  if (Res) { fat_close(V, Fd); return Res; }
  if (memcmp(".          ", D.Name, 11)) return FD32_ERROR_ACCESS_DENIED;
  Res = fat_read(V, Fd, (BYTE *) &D, 32, NULL);
  if (Res) { fat_close(V, Fd); return Res; }
  if (memcmp("..         ", D.Name, 11)) return FD32_ERROR_ACCESS_DENIED;
  /* And finally check that the following entries are free */
  do
  {
    Res = fat_read(V, Fd, (BYTE *) &D, 32, NULL);
    if (Res) { fat_close(V, Fd); return Res; }
    if ((D.Name[0] != FAT_DIR_FREE) && (D.Name[0] != FAT_DIR_ENDOFDIR))
      return FD32_ERROR_ACCESS_DENIED; /* Not empty */
  }
  while (D.Name[0] != FAT_DIR_ENDOFDIR);
  /* If we arrive here, it's Ok, we can unlink the directory */
  Res = fat_close(V, Fd);
  if (Res) return Res;
  return fat_unlink(V, DirName, FD32_ATTR_DIRECTORY, UseLfn);
}


/* Creates a new directory, preparing the "." and ".." entries and */
/* filling the rest of the first sector with nulls.                */
/* Returns 0 on success, or a negative error code on failure.      */
/* This is a public driver function.                               */
int fat_mkdir(void *P, char *DirName, int UseLfn)
{
  tFatFile     *F;
  tFatVolume   *V = (tFatVolume *) P;
  tFatDirEntry  D;
  int           Res;
  int           Fd;
  
  Fd = fat_open(V, DirName, FD32_OPEN_ACCESS_READWRITE |
                FAT_OPEN_DIRECTORY | FD32_OPEN_NOTEXIST_CREATE,
                FD32_ATTR_DIRECTORY, 0, NULL, UseLfn);
  if (Fd < 0) return Fd;
  F = &V->Files[Fd];

  /* Set to zero the whole cluster and rewind */
  #ifdef DEBUG
  fd32_log_printf("Zeroing cluster\n");
  #endif
  Res = fat_write(V, Fd, NULL,
                  V->Bpb.BPB_BytsPerSec * V->Bpb.BPB_SecPerClus, NULL);
  if (Res) { fat_close(V, Fd); return Res; }
  F->TargetPos = 0;
    
  /* Prepare the "dot" entry */
  #ifdef DEBUG
  fd32_log_printf("Writing dot entry\n");
  #endif
  D = F->DirEntry;
  memcpy(D.Name,".          ",11);
  Res = fat_write(V, Fd, (BYTE *) &D, 32, NULL);
  if (Res) { fat_close(V, Fd); return Res; }
  
  /* Prepare the "dotdot" entry */
  #ifdef DEBUG
  fd32_log_printf("Writing dotdot entry\n");
  #endif
  D.FstClusHI = F->ParentDirEntry.FstClusHI;
  D.FstClusLO = F->ParentDirEntry.FstClusLO;
  memcpy(D.Name,"..         ",11);
  Res = fat_write(V, Fd, (BYTE *) &D, 32, NULL);
  if (Res) { fat_close(V, Fd); return Res; }

  #ifdef DEBUG
  fd32_log_printf("Closing the directory\n");
  #endif
  return fat_close(V, Fd);
}


int fat_chdir(unsigned char *DirName, tFatVolume *V, int UseLfn)
{
  SDWORD Fd;
//  char   NewCurDir[FAT_MAX_LFN_PATH_LENGTH];

  Fd = fat_open(V, DirName, FD32_OPEN_ACCESS_READ | FAT_OPEN_DIRECTORY,
                FD32_ATTR_NONE, 0, NULL, UseLfn);
  if (Fd < 0) return (int) Fd;
  V->CurDirCluster = V->Files[Fd].FirstCluster;
//  fat_truename(NewCurDir, DirName, Vol)
//  for (Res = 0; (CurDirPath[Res] = NewCurDir[Res]); ++Res);
  fat_close(V, Fd);
  return 0;
}


int fat_cwd(unsigned char *Dest, tFatVolume *Vol)
{
  unsigned char *s = Vol->CurDirPath;
  for (; ((*Dest = *s) != 0); ++Dest, ++s);
  return 0;
}

