      /**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001-2002, Salvatore Isaja                               *
 *                                                                        *
 * This is "lfn.c" - Convert long file names in the standard DOS          *
 *                   directory entry format and generate short name       *
 *                   aliases from long file names                         *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/

#include "fat1.h"
#include "unicode.h"


/* The following codes have been extracted from the article:           */
/* "NOTES ON THE STRUCTURE OF THE VFAT FILESYSTEM", VFATInfo.txt,      */
/* by Galen C. Hunt <gchunt@cs.rochester.edu>, Gordon Chaffee,         */
/* from the table "NOTES ON UNICODE TRANSLATION IN VFAT FILESYSTEM"    */
/* by Steve Searle <steve@mgu.bath.ac.uk>.                             */
/*                                                                     */
/* Thanks to RHIDE macros which helped a lot in writing this in C! :-) */

/* Char used in short name instead of ASCII chars 128..255 */
static BYTE ShortNameEqu[128] =
{
  0x80, 0x9A, 0x90, 0xB6, 0x8E, 0xB7, 0x8F, 0x80,
  0xD2, 0xD3, 0xD4, 0xD8, 0xD7, 0xDE, 0x8E, 0x8F,
  0x90, 0x92, 0x92, 0xE2, 0x99, 0xE3, 0xEA, 0xEB,
  '_' , 0x99, 0x9A, 0x9D, 0x9C, 0x9D, 0x9E, 0x9F,
  0xB5, 0xD6, 0xE0, 0xE9, 0xA5, 0xA5, 0xA6, 0xA7,
  0xA8, 0xA9, 0xAA, 0xAB, 0xAC, 0xAD, 0xAE, 0xAF,
  0xB0, 0xB1, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7,
  0xB8, 0xB9, 0xBA, 0xBB, 0xBC, 0xBD, 0xBE, 0xBF,
  0xC0, 0xC1, 0xC2, 0xC3, 0xC4, 0xC5, 0xC7, 0xC7,
  0xC8, 0xC9, 0xCA, 0xCB, 0xCC, 0xCD, 0xCE, 0xCF,
  0xD1, 0xD1, 0xD2, 0xD3, 0xD4, 0x49, 0xD6, 0xD7,
  0xD8, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE, 0xDF,
  0xE0, 0xE1, 0xE2, 0xE3, 0x05, 0x05, 0xE6, 0xE8,
  0xE8, 0xE9, 0xEA, 0xEB, 0xED, 0xED, 0xEE, 0xEF,
  0xF0, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7,
  0xF8, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF
};


/* Calculate the 8-bit checksum for the long file name from its */
/* corresponding short name.                                    */
/* Called by split_lfn and find (find.c).                       */
BYTE lfn_checksum(tFatDirEntry *D)
{
  int Sum = 0, i;
  for (i=0;i<11;i++)
  {
    Sum = (((Sum & 1) << 7) | ((Sum & 0xFE) >> 1)) + D->Name[i];
  }
  return Sum;
}


/* Checks if a character Ch is valid for a FAT long file name.       */
/* Returns the valid character if successful, or 0 if Ch is invalid. */
/* Called by split_lfn.                                              */
static WORD validate_lfn_character(BYTE Ch)
{
  /* The following characters are not valid */
  if (Ch < 0x20) return 0;
  switch (Ch)
  {
    /* + , ; = [ ] . are not valid for short names but are ok for LFN */
    case 0x22 : /* " */
    case 0x2A : /* * */
    case 0x2F : /* / */
    case 0x3A : /* : */
    case 0x3C : /* < */
    case 0x3E : /* > */
    case 0x3F : /* ? */
    case 0x5C : /* \ */
    case 0x7C : /* | */ return 0;
  }
  return convert_oem_char_to_unicode(Ch);
}


/* Copies the 16-bit character Ch into the LFN slot Slot in the position */
/* SlotPos, taking care of the slot structure.                           */
/* Called by split_lfn.                                                  */
static void copy_char_in_lfn_slot(tFatLfnEntry *Slot, int SlotPos, WORD Ch)
{
  if ((SlotPos >= 0)  && (SlotPos <= 4))  Slot->Name0_4  [SlotPos]    = Ch;
  else
  if ((SlotPos >= 5)  && (SlotPos <= 10)) Slot->Name5_10 [SlotPos-5]  = Ch;
  else
  if ((SlotPos >= 11) && (SlotPos <= 12)) Slot->Name11_12[SlotPos-11] = Ch;
}


/* Splits the long file name in the string LongName to fit in up to 20 */
/* LFN slots, using the short name of the dir entry D to compute the   */
/* short name checksum.                                                */
/* On success, returns 0 and fills the Slot array with LFN entries and */
/* NumSlots with the number of entries occupied.                       */
/* Returns FAT_ERROR_INVALID_FILENAME on failure.                      */
/* Called by allocate_lfn_dir_entries (direntry.c).                    */
int split_lfn(tFatLfnEntry *Slot, tFatDirEntry *D, char *LongName, int *NumSlots)
{
  int  NamePos = 0;
  int  SlotPos = 0;
  int  Order   = 0;
  WORD Ch;

  /* Initialize the first slot */
  Slot[0].Order    = 1;
  Slot[0].Attr     = FD32_ATTR_LONGNAME;
  Slot[0].Reserved = 0;
  Slot[0].Checksum = lfn_checksum(D);
  Slot[0].FstClus  = 0;
  
  while (LongName[NamePos])
  {
    if (SlotPos == 13)
    {
      Order++;
      SlotPos = 0;
      Slot[Order].Order    = Order + 1; /* 1-based enumeration */
      Slot[Order].Attr     = FD32_ATTR_LONGNAME;
      Slot[Order].Reserved = 0;
      Slot[Order].Checksum = lfn_checksum(D);
      Slot[Order].FstClus  = 0;
    }
    
    Ch = validate_lfn_character(LongName[NamePos++]);
    if (Ch == 0) return FAT_RET_INVALID_FILENAME;
    if (NamePos == FD32_MAX_LFN_LENGTH) return FAT_RET_INVALID_FILENAME;

    copy_char_in_lfn_slot(&Slot[Order], SlotPos++, Ch);
  }
  /* Mark the slot as last */
  Slot[Order].Order |= 0x40;
  /* Insert an Unicode NULL terminator, only if it fits into the slots */
  copy_char_in_lfn_slot(&Slot[Order], SlotPos++, 0x0000);
  /* Pad the remaining characters of the slot with Unicode FFFFh */
  for (; SlotPos < 13; SlotPos++)
    copy_char_in_lfn_slot(&Slot[Order], SlotPos, 0xFFFF);

  *NumSlots = Order + 1;
  return 0;
}


/* Convert any (if any) invalid short name character of the long file    */
/* name starting at address From and ending at address To, putting the   */
/* result in the string Dest, long up to MaxLen characters.              */
/*                                                                       */
/* Returns:                                                              */
/*  0 - LFN is a 100% valid 8.3 file name, no LFN entry is needed        */
/*  1 - LFN fits into an 8.3 file name, but LFN entry is needed for case */
/*  2 - LFN doesn't fit into an 8.3 name, a "~Counter" will be added     */
/* Called by fat_generate_short_name.                                    */
static int convert_invalid_characters(unsigned char *Dest,
                                      unsigned char *From,
                                      unsigned char *To,
                                      int MaxLen)
{
  int k, Res = 0;
  for (k = 0; k < MaxLen;)
  {
    if (From == To)
    {
      Dest[k] = 0;
      return Res;
    }
    switch (*From)
    {
      case 0           : Dest[k] = 0; return Res;
      case ' '         :
      case '.'         : From++; Res = 2; break;
      case ':'         :
      case '+'         :
      case ','         :
      case ';'         :
      case '='         :
      case 152         :
      case '['         :
      case ']'         : Dest[k++] = '_'; From++; Res = 2; break;
      default          :
        if( (*From >= 'a') && (*From <= 'z') ) {
          Dest[k++] = *(From++) + 'A' - 'a'; Res = 1;
        } else if( ((*From >= 128) && (*From <= 151)) ||
                   ((*From >= 153) && (*From <= 255)) ) {
          Dest[k++] = ShortNameEqu[*(From++) - 128]; Res = 1;
        } else {
          Dest[k++] = *(From++);
        }
        break;
    }
  }
  Dest[k] = 0;
  if ((From == To) || (*From == 0)) return Res;
                               else return 2;
}


/* Generate a valid unique 8.3 file name UniqueShortName for the long  */
/* file name LongName in the directory Path of the FAT volume Vol.     */
/* If UniqueShortName == "" the LFN is intended to be an already       */
/* valid 8.3 name, so no LFN entires should be allocated for the file. */
/* Returns 0 on success.                                               */
/* Called by allocate_lfn_dir_entries (direntry.c).                    */
int fat_generate_short_name(tFatVolume *Vol, char *Path, char *LongName,
                            char *UniqueShortName, WORD Hint)
{
  char     *Ext, *s;
  char      ShortName[9],ShortExt[4],AsciiCounter[6];
  int       k, Fd, Counter = Hint, AsciiCounterLen, Res1, Res2, Res3 = 0;
  tFatFile *Dir;

  /* If there are initial periods remove them and mark LFN as not valid 8.3 */
  for (; *LongName == '.'; LongName++, Res3 = 2);
  /* Move to the end of the LongName string and remove trailing dots */
  for (Ext = LongName; *Ext != 0; Ext++);
  for (--Ext; *Ext == '.'; Ext--);
  /* Now move backward until '.' is found, if present */
  for (; (*Ext != '.') && (Ext != LongName); Ext--);
  if (Ext == LongName) Ext = 0; else Ext++;
  
  /* At this point LongName points to a null-terminated string containing */
  /* the long name, and Ext points to a null-terminated string containing */
  /* the long extension.                                                  */
  /* Now generate the valid short name (6 chars) and extension (3 chars). */
  Res1 = convert_invalid_characters(ShortName, LongName, Ext-1, 8);
  Res2 = convert_invalid_characters(ShortExt,  Ext,      NULL,  3);
//  printf("\nShortName: #%s#%i\nShortExt: #%s#%i\n",
//                  ShortName, Res1, ShortExt, Res2);

  /* If the LFN is a valid 8.3 file name, we don't need the LFN entry */
  if ((Res1 == 0) && (Res2 == 0) && (Res3 == 0))
  {
    UniqueShortName[0] = 0;
    return 0;
  }
  /* If the LFN doesn't fit into a 8.3 name, we need to add a "~Counter" */
  else if ((Res1 == 2) || (Res2 == 2) || (Res3 == 2))
  {
    Fd = fat_open(Vol, Path, FD32_OPEN_ACCESS_READ | FAT_OPEN_DIRECTORY,
                  FD32_ATTR_NONE, 0, NULL, FD32_LFN);
    if (Fd < 0) return Fd;
    Dir = &Vol->Files[Fd];

    /* Now append "~Counter" to the short name, incrementing "Counter" */
    /* until the file name is unique in that Path                      */
    for (;;)
    {
      /* Copy the name in the FullShortName and stop at the name's end */
      for (k = 0, s = ShortName; *s != 0; s++, k++) UniqueShortName[k] = *s;
      /* Go backward to fit the "~Counter" into 8 characters */
      itoa(Counter, AsciiCounter, 10);
      AsciiCounterLen = strlen(AsciiCounter);
      for (; k + AsciiCounterLen + 1 > 8; k--);
      /* Append the "~Counter" to the name */
      UniqueShortName[k++] = '~';
      for (s = AsciiCounter; *s != 0; s++, k++) UniqueShortName[k] = *s;
      /* And finally append the extension */
        UniqueShortName[k++] = '.';
      for (s = ShortExt; *s != 0; s++, k++) UniqueShortName[k] = *s;
      UniqueShortName[k] = 0;

//      printf("\nUniqueShortName: #%s#\n", UniqueShortName);

      /* Search for the generated name */
      Dir->TargetPos = 0;
      Res1 = find(Vol, Fd, UniqueShortName,
                  FD32_FIND_REQUIRE_NONE | FD32_FIND_ALLOW_ALL, NULL,
                  FD32_NOLFN);
      switch (Res1)
      {
        case 0                        : break;    /* No, that name exists   */
        case FD32_ERROR_NO_MORE_FILES : fat_close(Vol, Fd);
                                        return 0; /* Ok, the name is unique */
        default                       : fat_close(Vol, Fd);
                                        return Res1; /* A problem occurred  */
      }
      Counter++;
    }
  }
  /* If the LFN fits into an 8.3 name, but it's not in upper case, we */
  /* have to create an LFN entry for preserving the case and an 8.3   */
  /* upcased entry.                                                   */
  else
  {
    /* Copy the name in the FullShortName and stop at the name's end */
    for (k = 0, s = ShortName; *s != 0; s++, k++) UniqueShortName[k] = *s;
    /* And finally append the extension */
      UniqueShortName[k++] = '.';
    for (s = ShortExt; *s != 0; s++, k++) UniqueShortName[k] = *s;
    UniqueShortName[k] = 0;

//    printf("\nUniqueShortName: #%s#\n",UniqueShortName);
    return 0;
  }
}
