/**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001-2002, Salvatore Isaja                               *
 *                                                                        *
 * This is "read.c" - Read a block of data from an open file              *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/

#include "fat1.h"

/* Reads at most Size bytes from file F into Buffer.                 */
/* It may read less bytes than requested if End Of File is reached.  */
/* On success, returns 0 and, if Result is not NULL, puts the actual */
/* number of bytes read (zero meaning End Of File) in Result.        */
/* Returns a negative error code on failure.                         */
/* This is a public driver function.                                 */
int fat_read(void *P, int Fd, BYTE *Buffer, DWORD Size, DWORD *Result)
{
  DWORD        k;
  SDWORD       Res;
  int          NumBuf;
  tFatVolume  *V = (tFatVolume *) P;
  tFatFile    *F = validate_fd(V, Fd);
  unsigned int SectorOffset;
  long         MaxReadSize;

  /* If some error will happen, we'll return 0 bytes read */
  if (Result) *Result = 0;
  if (F == NULL) return FD32_ERROR_INVALID_HANDLE;

  /* Check if reading from file is allowed */
  if (((F->Mode & FD32_OPEN_ACCESS_MASK) != FD32_OPEN_ACCESS_READ)
   && ((F->Mode & FD32_OPEN_ACCESS_MASK) != FD32_OPEN_ACCESS_READWRITE))
    return FD32_ERROR_ACCESS_DENIED;

  /* Tell to move_to_targetpos that we are going to read */
  F->Mode &= ~FAT_FMODE_DO_WRITE;

  /* [KG] Check if we a reading from a File. */
  if( !(F->Mode & FAT_FMODE_ROOT) &&
      !(F->Mode & FAT_OPEN_DIRECTORY) ) {
    /* Yes => Adjust "Size" to maximum readable byte count */
    MaxReadSize = F->DirEntry.FileSize-F->TargetPos;
    if (MaxReadSize <= 0) {
      return 0;
    }
    else if (MaxReadSize < Size) {
      Size = MaxReadSize;
    }
  }

  for (k=0;k<Size;k++)
  {
    Res = move_to_targetpos(V, Fd, 0);
    if (Res)
    {
      if (Res == FAT_RET_EOF)
      {
        if (Result) *Result = k;
        if (((F->Mode & FD32_OPEN_ACCESS_MASK) != FD32_OPEN_ACCESS_READ_NA)
         && (!(F->Mode & FAT_OPEN_DIRECTORY)))
          update_date_and_time(NULL, &F->DirEntry.LstAccDate, NULL);
        return 0;
      }
      else
      {
        if (Result) *Result = k;
        return Res;
      }
    }
    /* Then we load the sector we sought and we read the byte we want */
    if (F->Mode & FAT_FMODE_ROOT)
      NumBuf = read_from_buffer(V, F->SectorInCluster +
                                first_root_dir_sector(&V->Bpb), &SectorOffset);
     else
      NumBuf = read_from_buffer(V, F->SectorInCluster +
                                first_sector_of_cluster(F->Cluster,V), &SectorOffset);
    if (NumBuf < 0) return NumBuf;
//[KG] SPEEDUP: transfer more than one byte, if possible...
//Noch zu bertragende Bytes: Size-k
//Im Puffer Verfgbare Bytes: 512-(F->ByteInSector)
    unsigned long TransferSize, LeftBytesInBuffer;
    TransferSize = Size-k;
    LeftBytesInBuffer = 512-F->ByteInSector; //(512*V->SectorsInBuffer)-F->ByteInSector-SectorOffset;
    if( TransferSize > LeftBytesInBuffer ) {
      TransferSize = LeftBytesInBuffer;
    }
    memcpy(&Buffer[k], &(V->Buffer[NumBuf].Data[F->ByteInSector+SectorOffset]), TransferSize);
//    Buffer[k] = V->Buffer[NumBuf].Data[F->ByteInSector];
    /* Finally we advance the file position */
    F->TargetPos += TransferSize;
    k += TransferSize-1;
  }
  /* Successful exit, whole Buffer read */
  if (Result) *Result = k;
  if (((F->Mode & FD32_OPEN_ACCESS_MASK) != FD32_OPEN_ACCESS_READ_NA)
   && (!(F->Mode & FAT_OPEN_DIRECTORY)))
    update_date_and_time(NULL, &F->DirEntry.LstAccDate, NULL);
  return 0;
}
