/**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001-2002, Salvatore Isaja                               *
 *                                                                        *
 * This is "support.c" - Small generic support functions                  *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/

#include "fat1.h"

unsigned long Timer;
static void interrupt (*oldaf_timer)(...);

void interrupt af_timer(...)
{

  Timer+=4;

  oldaf_timer;

}

void StartTimer(void)
{
  Timer = 0;
  oldaf_timer = getvect(0xaf);

  setvect(0xaf, af_timer);

  Timer = 0;

}

void StopTimer(void)

{

  setvect(0xaf, oldaf_timer);

  printf("Timer: %lu ms\n", Timer);

}


/* Gets the current date and time in standard DOS format:                   */
/*  Time bits: 15-11 hours (0-23), 10-5 minutes, 4-0 seconds/2              */
/*  Date bits: 15-9 year-1980, 8-5 month, 4-0 day                           */
/*  Hund: hundredths of second past the time in Time (0-199).               */
/* Pointers can be NULL.                                                    */
/* Called by fat_creat (open.c), fat_read (read.c) and fat_write (write.c). */
void update_date_and_time(WORD *Time, WORD *Date, BYTE *Hund)
{
  fd32_date_t CurrentDate;
  fd32_time_t CurrentTime;
  
  fd32_get_date(&CurrentDate);
  fd32_get_time(&CurrentTime);

  if (Hund) *Hund = CurrentTime.Hund + CurrentTime.Sec % 2 * 100;
  if (Time) *Time = ((CurrentTime.Sec / 2) & 0x1F)
                  + ((((unsigned long)CurrentTime.Min)  & 0x3F) << 5)
                  + ((((unsigned long)CurrentTime.Hour) & 0x1F) << 11);
  if (Date) *Date =  (CurrentDate.Day & 0x1F)
                  + ((unsigned long)(CurrentDate.Mon & 0x0F) << 5)
                  + ((unsigned long)((CurrentDate.Year - 1980) & 0xFF) << 9);
}


/* Checks if a file descriptor is into the range of valid file descriptors */
/* for a FAT volume and if the file is open.                               */
/* Returns a pointer to the file structure on success, or NULL on failure. */
/* Called by several dogs and pigs inside the driver.                      */
tFatFile *validate_fd(tFatVolume *V, DWORD Fd)
{
  if ((Fd < 0) || (Fd >= FAT_MAX_FILES)) return NULL;
  if (!(V->Files[Fd].Mode & FAT_FMODE_OPENED)) return NULL;
  return &V->Files[Fd];
}


/* The amount of clusters required to store N bytes */
/* Called by block_is_too_large (write.c).          */
DWORD clusters_amount(DWORD N, tFat32_Bpb *Bpb)
{
  return (N + Bpb->BPB_BytsPerSec * Bpb->BPB_SecPerClus - 1)
           / (Bpb->BPB_BytsPerSec * Bpb->BPB_SecPerClus);
}


#if 0
/* The bytes actually allocated for the file F (usually > FileSize) */
static inline DWORD file_occupation_in_bytes(tFatFile *F)
{
  return clusters_amount(F->FileSize, &F->Volume->Bpb)
         * F->Volume->Bpb.BPB_BytsPerSec
         * F->Volume->Bpb.BPB_SecPerClus;
}
#endif


int fat_free_disk_space(tFatVolume *V, DWORD *SectorsPerCluster,
                        DWORD *FreeClusters,
                        DWORD *BytesPerSector,
                        DWORD *TotalClusters)
{
  if (SectorsPerCluster) *SectorsPerCluster = V->Bpb.BPB_SecPerClus;
  if (FreeClusters)      *FreeClusters      = V->FSI_Free_Count;
  if (BytesPerSector)    *BytesPerSector    = V->Bpb.BPB_BytsPerSec;
  if (TotalClusters)     *TotalClusters     = V->DataClusters;
  return 0;
}


/* The starting sector of the root directory of a FAT12/FAT16 volume.  */
/* For FAT32 volumes, the root directory is just a file like all other */
/* directories, starting at cluster BPB_RootClus.                      */
/* Called by fat_read (read.c) and fat_write (write.c).                */
DWORD first_root_dir_sector(tFat32_Bpb *Bpb)
{
  return Bpb->BPB_ResvdSecCnt + (Bpb->BPB_NumFATs * Bpb->BPB_FATSz16);
}


/* The sector number of the first sector of a valid data cluster N, where: */
/* - the first valid data cluster is 2 (cluster 0 and 1 are reserved)      */
/* - the last valid data cluster is Vol.DataClusters + 1                   */
/* - the number of clusters (including reserved) is Vol.DataClusters + 2   */
/* Returns 0 if the N is out of range of valid clusters.                   */
/* Called by fat_read (read.c) and fat_write (write.c).                    */
DWORD first_sector_of_cluster(DWORD N, tFatVolume *Vol)
{
  if ((N < 2) || (N > Vol->DataClusters + 1)) return 0;
  return ((N - 2) * Vol->Bpb.BPB_SecPerClus) + Vol->FirstDataSector;
}
