/**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001, Salvatore Isaja                                    *
 *                                                                        *
 * This is "wildcard.c" - Compare two ASCII or Unicode strings with       *
 *                        wildcards (* and ?) allowed in one of them      *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/

#include "unicode.h"

#define UNICODE_DOT      0x002E
#define UNICODE_ASTERISK 0x002A
#define UNICODE_QUESTION 0x003F


int stricmp_with_wildcards(unsigned char *s1, unsigned char *s2)
{
  /* Long File Names without periods have an implicit trailing period
   * that has to be matched in searches like *.*
   */
  unsigned char *s;
  int           ImplicitDot = 1;
  for (s = s1; *s; s++) if (*s == '.') ImplicitDot = 0;
  
  do
  {
    /* If there is an '*' wildcard, we have to perform a variable-position
     * pattern matching.
     */
    if (*s2 == '*')
    {
      int           Found;
      unsigned char *s1Old = s1,
                    *s2Old = ++s2;
      /* If this '*' was the last wildcard, we don't need to scan further. */
      if (*s2 == 0) return 0;
      /* Search for the s2's characters of before another '*' (or end of s2),
       * including the null terminator. If we don't find that pattern in s1,
       * strcmp fails. If we find that pattern, s1 and s2 are advanced beyond
       * the pattern and the comparison restarts from there.
       */
      do
      {
        Found = 1;
        s2    = s2Old;
        while (*s2 != '*')
        {
          if ((*s1 == 0) && ImplicitDot)
          {
            ImplicitDot = 0;
            if ((*s2 != '?') && ('.' != *s2) && *s2)
            {
              Found = 0;
              break;
            }
            if (*s2 == 0) break;
            s2++;
          }
          else
          {
            if ((*s2 != '?') || (*s1 == 0))
              if (toupper(*s1) != toupper(*s2))
              {
                Found = 0;
                break;
              }
            if (*s2 == 0) break;
            s1++;
            s2++;
          }
        }
        if (Found) break;
        if (*s1Old) s1 = ++s1Old;
      }
      while (*s1 || ImplicitDot);
      if (!Found) return 1;
    }
    else
    /* If an '*' is not there, we have simply to compare char by char */
    {
      if ((*s1 == 0) && ImplicitDot)
      {
        ImplicitDot = 0;
        if ((*s2 != '?') && ('.' != toupper(*s2))) return 1;
        s2++;
      }
      else
      {
        if ((*s2 != '?') || (*s1 == 0))
          if (toupper(*s1) != toupper(*s2))
            return 1;
        s1++;
        s2++;
      }
    }
  }
  while (*s2 || *s1);
  return 0;
}


int unicode_stricmp_with_wildcards(unsigned short *s1, unsigned short *s2)
{
  /* Long File Names without periods have an implicit trailing period
   * that has to be matched in searches like *.*
   */
  unsigned short *s;
  int            ImplicitDot = 1;
  for (s = s1; *s; s++) if (*s == UNICODE_DOT) ImplicitDot = 0;
  
  do
  {
    /* If there is an '*' wildcard, we have to perform a variable-position
     * pattern matching.
     */
    if (*s2 == UNICODE_ASTERISK)
    {
      int            Found;
      unsigned short *s1Old = s1,
                     *s2Old = ++s2;
      /* If this '*' was the last wildcard, we don't need to scan further. */
      if (*s2 == 0) return 0;
      /* Search for the s2's characters of before another '*' (or end of s2),
       * including the null terminator. If we don't find that pattern in s1,
       * strcmp fails. If we find that pattern, s1 and s2 are advanced beyond
       * the pattern and the comparison restarts from there.
       */
      do
      {
        Found = 1;
        s2    = s2Old;
        while (*s2 != UNICODE_ASTERISK)
        {
          if ((*s1 == 0) && ImplicitDot)
          {
            ImplicitDot = 0;
            if ((*s2 != UNICODE_QUESTION) && (UNICODE_DOT != *s2) && *s2)
            {
              Found = 0;
              break;
            }
            if (*s2 == 0) break;
            s2++;
          }
          else
          {
            if ((*s2 != UNICODE_QUESTION) || (*s1 == 0))
              if (unicode_toupper(*s1) != unicode_toupper(*s2))
              {
                Found = 0;
                break;
              }
            if (*s2 == 0) break;
            s1++;
            s2++;
          }
        }
        if (Found) break;
        if (*s1Old) s1 = ++s1Old;
      }
      while (*s1 || ImplicitDot);
      if (!Found) return 1;
    }
    else
    /* If an '*' is not there, we have simply to compare char by char */
    {
      if ((*s1 == 0) && ImplicitDot)
      {
        ImplicitDot = 0;
        if ((*s2 != UNICODE_QUESTION) && (UNICODE_DOT != *s2)) return 1;
        s2++;
      }
      else
      {
        if ((*s2 != UNICODE_QUESTION) || (*s1 == 0))
          if (unicode_toupper(*s1) != unicode_toupper(*s2))
            return 1;
        s1++;
        s2++;
      }
    }
  }
  while (*s2 || *s1);
  return 0;
}


int utf8_stricmp_with_wildcards(char *s1, char *s2)
{
  /* Long File Names without periods have an implicit trailing period
   * that has to be matched in searches like *.*
   */
  char *s;
  int            ImplicitDot = 1;
  for (s = s1; *s; s++) if (*s == '.') ImplicitDot = 0;
  
  do
  {
    /* If there is an '*' wildcard, we have to perform a variable-position
     * pattern matching.
     */
    if (*s2 == '*')
    {
      int   Found;
      char *s1Old = s1,
           *s2Old = ++s2;
      /* If this '*' was the last wildcard, we don't need to scan further. */
      if (*s2 == 0) return 0;
      /* Search for the s2's characters of before another '*' (or end of s2),
       * including the null terminator. If we don't find that pattern in s1,
       * strcmp fails. If we find that pattern, s1 and s2 are advanced beyond
       * the pattern and the comparison restarts from there.
       */
      do
      {
        Found = 1;
        s2    = s2Old;
        while (*s2 != UNICODE_ASTERISK)
        {
          if ((*s1 == 0) && ImplicitDot)
          {
            ImplicitDot = 0;
            if ((*s2 != '?') && ('.' != *s2) && *s2)
            {
              Found = 0;
              break;
            }
            if (*s2 == 0) break;
            s2++;
          }
          else
          {
            if ((*s2 != '?') || (*s1 == 0))
              if (unicode_toupper(*s1) != unicode_toupper(*s2))
              {
                Found = 0;
                break;
              }
            if (*s2 == 0) break;
            s1++;
            s2++;
          }
        }
        if (Found) break;
        if (*s1Old) s1 = ++s1Old;
      }
      while (*s1 || ImplicitDot);
      if (!Found) return 1;
    }
    else
    /* If an '*' is not there, we have simply to compare char by char */
    {
      if ((*s1 == 0) && ImplicitDot)
      {
        ImplicitDot = 0;
        if ((*s2 != '?') && ('.' != *s2)) return 1;
        s2++;
      }
      else
      {
        if ((*s2 != '?') || (*s1 == 0))
          if (unicode_toupper(*s1) != unicode_toupper(*s2))
            return 1;
        s1++;
        s2++;
      }
    }
  }
  while (*s2 || *s1);
  return 0;
}

